# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""This module contains objects for low-level representation of lattice systems."""

import logging
from copy import deepcopy
from typing import Union, Sequence
import numpy as np
from scipy.sparse import csr_matrix, bsr_matrix
from .utils import create_lookup_table, min_dtype

__all__ = ["DataMap", "LatticeData"]

logging.captureWarnings(True)

logger = logging.getLogger(__name__)


class DataMap:
    """Object for low-level representation of sites and site-pairs.

    Parameters
    ----------
    alphas : (N) np.ndarray
        The atom indices of the sites.
    pairs : (M, 2) np.ndarray
        An array of index-pairs of the lattice sites.
    distindices : (M) np.ndarray
        The distance-indices for each pair

    Notes
    -----
    This object is not intended to be instantiated by the user. Use the ``map`` method
    of ``latticeData`` or the ``dmap`` method of the main ``Lattice```object.
    """

    def __init__(self, alphas: np.ndarray, pairs: np.ndarray, distindices: np.ndarray):
        sites = np.arange(len(alphas), dtype=pairs.dtype)
        map_ = np.append(-alphas - 1, distindices)
        indices_ = np.append(np.tile(sites, (2, 1)).T, pairs, axis=0)
        ind = np.argsort(indices_[:, 0])
        self._map = map_[ind]
        self._indices = indices_[ind]

    @property
    def size(self) -> int:
        """The number of the data points (sites + neighbor pairs)"""
        return len(self._indices)

    @property
    def indices(self) -> np.ndarray:
        """The indices of the data points as rows and collumns."""
        return self._indices.T

    @property
    def rows(self):
        """The rows of the data points."""
        return self._indices[:, 0]

    @property
    def cols(self):
        """The columns of the data points."""
        return self._indices[:, 1]

    @property
    def nbytes(self):
        """The number of bytes stored in the datamap."""
        return self._map.nbytes + self._indices.nbytes

    def indices_indptr(self):
        """Constructs the `indices` and `indptr` arrays used for CSR/BSR matrices.

        CSR/BSR sparse matrix format:
        The block column indices for row i are stored in
        ``indices[indptr[i]:indptr[i+1]]`` and their corresponding block values are
        stored in ``data[indptr[i]: indptr[i+1]]``.

        Returns
        -------
        indices : (N, ) np.ndarray
            CSR/BSR format index array.
        indptr : (M, ) np.ndarray
            CSR/BSR format index pointer array.

        See Also
        --------
        scipy.sparse.csr_matrix : Compressed Sparse Row matrix
        scipy.sparse.bsr_matrix : Block Sparse Row matrix
        """
        rows, cols = self._indices.T
        indptr, indices = [0], []
        unique = np.sort(np.unique(rows))
        for r in unique:
            mask = rows == r
            indices.extend(list(cols[mask]))
            indptr.append(len(indices))
        return indices, indptr

    def onsite(self, alpha: int = None) -> np.ndarray:
        """Creates a mask of the site elements for the atoms with the given index.

        Parameters
        ----------
        alpha : int, optional
            Index of the atom in the unitcell. If `None` a mask for all atoms
            is returned. The default is `None`.

        Returns
        -------
        mask : np.ndarray
        """
        if alpha is None:
            return self._map < 0
        return self._map == -alpha - 1

    def hopping(self, distidx: int = None) -> np.ndarray:
        """Creates a mask of the site-pair elements with the given distance index.

        Parameters
        ----------
        distidx : int, optional
            Index of distance to neighboring sites, default is 0 (nearest neighbors).
            If `None` a mask for neighbor-connections is returned. The default is
            `None`.

        Returns
        -------
        mask : np.ndarray
        """
        if distidx is None:
            return self._map >= 0
        return self._map == distidx

    def zeros(self, norb=None, dtype=None):
        """Creates an empty data-arary.

        Parameters
        ----------
        norb : int, optional
            The number of orbitals M. By default, only a single orbital is used.
        dtype : int or str or np.dtype, optional
            The data type of the array. By default, it is set automatically.

        Returns
        -------
        data : np.ndarray
            The empty data array. If a single orbital is used the array is
            one-dimensional, otherwise the array has the shape (N, M, M).
        """
        if norb is None:
            shape = self.size
        else:
            shape = (self.size, norb, norb)
        return np.zeros(shape, dtype=dtype)

    def build_csr(self, data, shape=None, dtype=None):
        """Constructs a CSR matrix using the given data and the indices of the data map.

        Parameters
        ----------
        data : (N,) np.ndarray
            The input data for constructing the CSR matrix. The data array should be
            filled using the built-in mask methods of the `DataMap` class.
        shape : tuple, optional
            The shape of the resulting matrix. If None (default), the shape is inferred
            from the data and indices of the matrix.
        dtype : int or str or np.dtype, optional
            The data type of the matrix. By default, it is set automatically.

        Returns
        -------
        sparse_mat : (M, M) scipy.sparse.csr.csr_matrix
            The sparse matrix representing the lattice data.
        """
        return csr_matrix((data, self.indices), shape=shape, dtype=dtype)

    def build_bsr(self, data, shape=None, dtype=None):
        """Constructs a BSR matrix using the given data and the indices of the data map.

        Parameters
        ----------
        data : (N, B, B) np.ndarray
            The input data for constructing the BSR matrix. The array must be
            3-dimensional, where the first axis N represents the number of blocks
            and the last two axis B the size of each block. The data array should be
            filled using the built-in mask methods of the `DataMap` class.
        shape : tuple, optional
            The shape of the resulting matrix. If None (default), the shape is inferred
            from the data and indices of the matrix.
        dtype : int or str or np.dtype, optional
            The data type of hte matrix. By default, it is set automatically.

        Returns
        -------
        sparse_mat : (M, M) scipy.sparse.csr.bsr_matrix
            The sparse matrix representing the lattice data.
        """
        return bsr_matrix((data, *self.indices_indptr()), shape=shape, dtype=dtype)

    def __repr__(self):
        return f"{self.__class__.__name__}(size: {self.size})"


class LatticeData:
    """Object for storing the indices, positions and neighbors of lattice sites.

    Parameters
    ----------
    indices : array_like of iterable of int
        The lattice indices of the sites.
    positions : array_like of iterable of int
        The positions of the sites.
    neighbors : iterable of iterable of of int
        The neighbors of the sites.
    distances : iterabe of iterable of int
        The distances of the neighbors.
    """

    def __init__(self, *args):
        self.indices = np.array([])
        self.positions = np.array([])
        self.neighbors = np.array([])
        self.distances = np.array([])
        self.distvals = np.array([])
        self.pnvecs = np.array([])
        self.pmask = np.array([])
        self.paxes = np.array([])

        self.invalid_idx = -1
        self.invalid_distidx = -1
        self._dmap = None

        if args:
            self.set(*args)

    @property
    def dim(self) -> int:
        """The dimension of the data points."""
        return self.positions.shape[1]

    @property
    def num_sites(self) -> int:
        """The number of sites stored."""
        return self.indices.shape[0]

    @property
    def num_distances(self) -> int:
        """The number of distances of the neighbor data."""
        return len(np.unique(self.distances[np.isfinite(self.distances)]))

    @property
    def nbytes(self):
        """Returns the number of bytes stored."""
        size = self.indices.nbytes + self.positions.nbytes
        size += self.neighbors.nbytes + self.distances.nbytes
        size += self.distvals.nbytes + self.paxes.nbytes
        size += self.pnvecs.nbytes + self.pmask.nbytes
        return size

    def copy(self) -> "LatticeData":
        """Creates a deep copy of the instance."""
        return deepcopy(self)

    def reset(self) -> None:
        """Resets the `LatticeData` instance."""
        self.indices = np.array([])
        self.positions = np.array([])
        self.neighbors = np.array([])
        self.distances = np.array([])
        self.distvals = np.array([])
        self.pnvecs = np.array([])
        self.paxes = np.array([])
        self.pmask = np.array([])
        self._dmap = None
        self.invalid_idx = -1
        self.invalid_distidx = -1

    def set(
        self,
        indices: np.ndarray,
        positions: np.ndarray,
        neighbors: np.ndarray,
        distances: np.ndarray,
    ) -> None:
        """Sets the data of the `LatticeData` instance.

        Parameters
        ----------
        indices : (N, D+1) np.ndarray
            The lattice indices of the sites.
        positions : (N, D) np.ndarray
            The positions of the sites.
        neighbors : (N, M) np.ndarray
            The neighbors of the sites.
        distances : (N, M) iterabe of iterable of int
            The distances of the neighbors.
        """
        logger.debug("Setting data")
        distvals, distidx = create_lookup_table(distances)

        self.indices = indices
        self.positions = positions
        self.neighbors = neighbors
        self.distances = distidx
        self.distvals = distvals
        self.paxes = np.zeros((*self.neighbors.shape, self.dim), dtype=np.int8)
        self.pnvecs = np.zeros((*self.neighbors.shape, self.dim), dtype=indices.dtype)
        self.pmask = np.zeros_like(self.neighbors, dtype=bool)

        self.invalid_idx = self.num_sites
        self.invalid_distidx = np.max(self.distances)
        self._dmap = None

    def remove(self, sites):
        sites = np.atleast_1d(sites)

        # store current invalid index
        invalid_idx = self.invalid_idx
        invalid_distidx = self.invalid_distidx

        # Remove data from arrays
        indices = np.delete(self.indices, sites, axis=0)
        positions = np.delete(self.positions, sites, axis=0)
        neighbors = np.delete(self.neighbors, sites, axis=0)
        distances = np.delete(self.distances, sites, axis=0)

        # Update neighbor indices and distances:
        # For each removed site below the neighbor index has to be decremented once
        mask = np.isin(neighbors, sites)
        neighbors[mask] = invalid_idx
        distances[mask] = invalid_distidx
        for count, i in enumerate(sorted(sites)):
            neighbors[neighbors > (i - count)] -= 1

        # Update invalid indices in neighbor array since number of sites changed
        num_sites = indices.shape[0]
        neighbors[neighbors == invalid_idx] = num_sites

        # Set updated data
        self.indices = indices
        self.positions = positions
        self.neighbors = neighbors
        self.distances = distances
        self.invalid_idx = num_sites

    def get_limits(self) -> np.ndarray:
        """Computes the geometric limits of the positions of the stored sites.

        Returns
        -------
        limits : np.ndarray
            The minimum and maximum value for each axis of the position data.
        """
        return np.array(
            [np.min(self.positions, axis=0), np.max(self.positions, axis=0)]
        )

    def get_index_limits(self) -> np.ndarray:
        """Computes the geometric limits of the lattice indices of the stored sites.

        Returns
        -------
        limits: np.ndarray
            The minimum and maximum value for each axis of the lattice indices.
        """
        return np.array([np.min(self.indices, axis=0), np.max(self.indices, axis=0)])

    def get_cell_limits(self) -> np.ndarray:
        """Computes the geometric limits of the lattice cells of the stored sites.

        Returns
        -------
        limits: np.ndarray
            The minimum and maximum value for each axis of the translation indices.
        """
        indices = self.indices[:, :-1]
        return np.array([np.min(indices, axis=0), np.max(indices, axis=0)])

    def get_translation_limits(self) -> np.ndarray:  # pragma: no cover
        """Computes the geometric limits of the translation vectors of the stored sites.

        Returns
        -------
        limits : np.ndarray
            The minimum and maximum value for each axis of the lattice indices.
        """
        return self.get_index_limits()[:, :-1]

    def neighbor_mask(
        self,
        site: int,
        distidx: int = None,
        periodic: bool = None,
        unique: bool = False,
    ) -> np.ndarray:
        """Creates a mask for the valid neighbors of a specific site.

        Parameters
        ----------
        site : int
            The index of the site.
        distidx : int, optional
            The index of the distance. If `None` the data for all distances is returned.
            The default is `None` (all neighbors).
        periodic : bool, optional
            Periodic neighbor flag. If `None` the data for all neighbors is returned.
            If a bool is passed either the periodic or non-periodic neighbors
            are masked. The default is `None` (all neighbors).
        unique : bool, optional
            If 'True', each unique pair is only return once. The defualt is `False`.

        Returns
        -------
        mask : np.ndarray
        """
        if distidx is None:
            mask = self.distances[site] < self.invalid_distidx
        else:
            mask = self.distances[site] == distidx
        if unique:
            mask &= self.neighbors[site] > site

        if periodic is not None:
            mask &= self.pmask[site] if periodic else ~self.pmask[site]
        return mask

    def set_periodic(
        self, indices: dict, distances: dict, nvecs: dict, axes: dict
    ) -> None:
        """Adds periodic neighbors to the invalid slots of the neighbor data

        Parameters
        ----------
        indices : dict
            Indices of the periodic neighbors. All dictionaries have the site as key
            and a list of ``np.ndarray`` as values.
        distances : dict
            The distances of the periodic neighbors.
        nvecs : dict
            The translation vectors of the periodic neighbors.
        axes : dict
            Index of the translation axis of the periodic neighbors.
        """
        # remove previous periodic neighbors
        self.remove_periodic()
        # Set new periodic neighbors
        for i, pidx in indices.items():
            # compute invalid slots of normal data
            i0 = len(self.get_neighbors(i, periodic=False))
            i1 = i0 + len(pidx)
            # translate distances to indices
            dists = distances[i]
            distidx = [np.searchsorted(self.distvals, d) for d in dists]
            # add periodic data
            self.neighbors[i, i0:i1] = pidx
            self.distances[i, i0:i1] = distidx
            self.paxes[i, i0:i1, : axes[i].shape[1]] = axes[i]
            self.pnvecs[i, i0:i1] = nvecs[i]
            self.pmask[i, i0:i1] = 1

    def sort(self, ax=None, indices=None, reverse=False):
        if ax is not None:
            indices = np.lexsort(self.indices.T[[ax]])
        if reverse:
            indices = indices[::-1]
        # Reorder data
        self.indices = self.indices[indices]
        self.positions = self.positions[indices]
        self.neighbors = self.neighbors[indices]
        self.distances = self.distances[indices]
        self.paxes = self.paxes[indices]
        self.pnvecs = self.pnvecs[indices]
        self.pmask = self.pmask[indices]

        # Translate neighbor indices
        old_neighbors = self.neighbors.copy()
        for new, old in enumerate(indices):
            mask = old_neighbors == old
            self.neighbors[mask] = new

    def remove_periodic(self):
        mask = self.pmask
        self.neighbors[mask] = self.invalid_idx
        self.distances[mask] = self.invalid_distidx
        self.pnvecs.fill(0)
        self.pmask.fill(0)
        self.paxes.fill(self.dim)

    def sort_neighbors(self):
        distances = self.distvals[self.distances]
        i = np.arange(len(distances))[:, np.newaxis]
        j = np.argsort(distances, axis=1)
        self.neighbors = self.neighbors[i, j]
        self.distances = self.distances[i, j]
        self.paxes = self.paxes[i, j]
        self.pnvecs = self.pnvecs[i, j]
        self.pmask = self.pmask[i, j]

    def add_neighbors(self, site, neighbors, distances):
        neighbors = np.atleast_2d(neighbors)
        distances = np.atleast_2d(distances)
        # compute invalid slots of normal data
        i0 = len(self.distances[site, self.distances[site] != self.invalid_distidx])
        i1 = i0 + len(neighbors)
        # Translate distances to indices
        distidx = np.asarray([np.searchsorted(self.distvals, d) for d in distances])
        # Add new neighbor data to unused slots
        self.neighbors[site, i0:i1] = neighbors
        self.distances[site, i0:i1] = distidx

    def append(self, *args, copy=False):
        neighbors1 = self.neighbors.copy()
        distances1 = self.distvals[self.distances]
        if len(args) == 1 and isinstance(args[0], LatticeData):
            data = args[0]
            indices2 = data.indices
            positions2 = data.positions
            neighbors2 = data.neighbors
            distances2 = data.distvals[data.distances]
        else:
            indices2, positions2, neighbors2, distances2 = args

        # Remove periodic neighbors
        self.remove_periodic()

        # Convert invalid indices of neighbor data
        invalid_idx = self.num_sites + len(indices2)
        neighbors1[neighbors1 == self.invalid_idx] = invalid_idx
        neighbors2[neighbors2 == len(indices2)] = invalid_idx

        # upgrade dtype if neccessary
        dtype = min_dtype(np.max(neighbors2) + self.num_sites)
        if dtype != neighbors2.dtype:
            neighbors2 = neighbors2.astype(dtype)
        # Shift neighbor indices
        neighbors2[neighbors2 != invalid_idx] += self.num_sites

        # Pad neighbor data
        cols1 = neighbors1.shape[1]
        cols2 = neighbors2.shape[1]
        cols = max(cols1, cols2)
        if cols1 < cols:
            widths = ((0, 0), (0, cols - cols1))
            neighbors1 = np.pad(neighbors1, widths, constant_values=invalid_idx)
            distances1 = np.pad(distances1, widths, constant_values=np.inf)
        if cols2 < cols:
            widths = ((0, 0), (0, cols - cols2))
            neighbors2 = np.pad(neighbors2, widths, constant_values=invalid_idx)
            distances2 = np.pad(distances2, widths, constant_values=np.inf)

        # Join data
        indices = np.append(self.indices, indices2, axis=0)
        positions = np.append(self.positions, positions2, axis=0)
        neighbors = np.append(neighbors1, neighbors2, axis=0)
        distances = np.append(distances1, distances2, axis=0)

        if copy:
            data = LatticeData(indices, positions, neighbors, distances)
            data.sort_neighbors()
            return data

        self.set(indices, positions, neighbors, distances)
        self.sort_neighbors()
        return self

    def get_positions(self, alpha):
        """Returns the atom positions of a sublattice."""
        mask = self.indices[:, -1] == alpha
        return self.positions[mask]

    def get_neighbors(
        self,
        site: int,
        distidx: int = None,
        periodic: bool = None,
        unique: bool = False,
    ) -> np.ndarray:
        """Returns the neighbors of a lattice site.

        See the `neighbor_mask`-method for more information on parameters

        Returns
        -------
        neighbors : np.ndarray
            The indices of the neighbors.
        """
        mask = self.neighbor_mask(site, distidx, periodic, unique)
        return self.neighbors[site, mask]

    def iter_neighbors(self, site: int, unique: bool = False) -> np.ndarray:
        """Iterates over the neighbors of all distance levels.

        See the `neighbor_mask`-method for more information on parameters

        Yields
        ------
        distidx : int
        neighbors : np.ndarray
        """
        for distidx in np.unique(self.distances[site]):
            if distidx != self.invalid_distidx:
                yield distidx, self.get_neighbors(site, distidx, unique=unique)

    def map(self) -> DataMap:
        """Builds a map containing the atom-indices, site-pairs and distances.

        Returns
        -------
        datamap : DataMap
        """
        if self._dmap is None:
            alphas = self.indices[:, -1].astype(np.int8)
            # Build index pairs and corresponding distance array
            dtype = np.min_scalar_type(self.num_sites)
            sites = np.arange(self.num_sites, dtype=dtype)
            sites_t = np.tile(sites, (self.neighbors.shape[1], 1)).T
            pairs = np.reshape([sites_t, self.neighbors], newshape=(2, -1)).T
            distindices = self.distances.flatten()
            # Filter pairs with invalid indices
            mask = distindices != self.invalid_distidx
            pairs = pairs[mask]
            distindices = distindices[mask]
            self._dmap = DataMap(alphas, pairs.astype(dtype), distindices)
        return self._dmap

    def site_mask(
        self,
        mins: Sequence[Union[float, None]] = None,
        maxs: Sequence[Union[float, None]] = None,
        invert: bool = False,
    ) -> np.ndarray:
        """Creates a mask for the position data of the sites.

        Parameters
        ----------
        mins : sequence or float or None, optional
            Optional lower bound for the positions. The default is no lower bound.
        maxs : sequence or float or None, optional
            Optional upper bound for the positions. The default is no upper bound.
        invert : bool, optional
            If `True`, the mask is inverted. The default is `False`.

        Returns
        -------
        mask : np.ndarray
            The mask containing a boolean value for each site.
        """
        if mins is None:
            mins = [None] * self.dim
        elif len(mins) != self.dim:
            mins = list(mins) + list([None] * (self.dim - len(mins)))
        if maxs is None:
            maxs = list([None] * self.dim)
        elif len(maxs) != self.dim:
            maxs = list(maxs) + [None] * (self.dim - len(maxs))
        mins = [(x if x is not None else -np.inf) for x in mins]
        maxs = [(x if x is not None else +np.inf) for x in maxs]
        limits = np.array([mins, maxs])
        mask = 1
        for ax in range(self.dim):
            ax_data = self.positions[:, ax]
            mask &= (limits[0, ax] <= ax_data) & (ax_data <= limits[1, ax])
        return np.asarray(1 - mask if invert else mask)

    def find_sites(
        self,
        mins: Sequence[Union[float, None]] = None,
        maxs: Sequence[Union[float, None]] = None,
        invert: bool = False,
    ) -> np.ndarray:
        """Returns the indices of sites inside or outside the given limits.

        Parameters
        ----------
        mins : sequence or float or None, optional
            Optional lower bound for the positions. The default is no lower bound.
        maxs : sequence or float or None, optional
            Optional upper bound for the positions. The default is no upper bound.
        invert : bool, optional
            If `True`, the mask is inverted and the positions outside of the bounds
            will be returned. The default is `False`.

        Returns
        -------
        indices: np.ndarray
            The indices of the masked sites.
        """
        mask = self.site_mask(mins, maxs, invert)
        return np.where(mask)[0]

    def find_outer_sites(self, ax: int, offset: int) -> np.ndarray:
        """Returns the indices of the outer sites along a specific axis.

        Parameters
        ----------
        ax : int
            The geometrical axis.
        offset : int
            The width of the outer slices.

        Returns
        -------
        indices : np.ndarray
            The indices of the masked sites.
        """
        limits = self.get_limits()
        mins = [None for _ in range(self.dim)]
        maxs = [None for _ in range(self.dim)]
        mins[ax] = limits[0, ax] + offset
        maxs[ax] = limits[1, ax] - offset
        mask = self.site_mask(mins, maxs, invert=True)
        return np.where(mask)[0]

    def __bool__(self) -> bool:  # pragma: no cover
        return bool(len(self.indices))

    def __str__(self) -> str:
        w = 9, 15, 10
        delim = " | "
        headers = "Indices", "Positions", "Neighbours"
        lines = list()
        s = f"{headers[0]:<{w[0]}}{delim}{headers[1]:<{w[1]}}{delim}{headers[2]}"
        lines.append(s)
        for site in range(self.num_sites):
            pos = "[" + ", ".join(f"{x:.1f}" for x in self.positions[site]) + "]"
            idx = str(self.indices[site])
            neighbors = str(self.neighbors[site])
            s = f"{idx:<{w[0]}}{delim}{pos:<{w[1]}}{delim}{neighbors}"
            lines.append(s)
        return "\n".join(lines)



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Tools for dispersion computation and plotting."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as tck
from .utils import chain
from .spatial import distance
from .plotting import draw_lines
from .atom import Atom

__all__ = [
    "bandpath_subplots",
    "plot_dispersion",
    "disp_dos_subplots",
    "plot_disp_dos",
    "plot_bands",
    "DispersionPath",
]


def _color_list(color, num_bands):
    if color is None:
        colors = [f"C{i}" for i in range(num_bands)]
    elif isinstance(color, str) or not hasattr(color, "__len__"):
        colors = [color] * num_bands
    else:
        colors = color
    return colors


def _scale_xaxis(num_points, disp, scales=None):
    sect_size = len(disp) / (num_points - 1)
    scales = np.ones(num_points - 1) if scales is None else scales
    k0, k, ticks = 0, [], [0]
    for scale in scales:
        k.extend(k0 + np.arange(sect_size) * scale)
        k0 = k[-1]
        ticks.append(k0)
    return k, ticks


def _set_piticks(axis, num_ticks=2, frmt=".1f"):
    axis.set_major_formatter(tck.FormatStrFormatter(rf"%{frmt} $\pi$"))
    axis.set_major_locator(tck.LinearLocator(2 * num_ticks + 1))


def bandpath_subplots(ticks, labels, xlabel="$k$", ylabel="$E(k)$", grid="both"):
    fig, ax = plt.subplots()
    ax.set_xlim(0, ticks[-1])
    ax.set_xticks(ticks)
    ax.set_xticklabels(labels)
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    if grid:
        if not isinstance(grid, str):
            grid = "both"
        ax.set_axisbelow(True)
        ax.grid(b=True, which="major", axis=grid)
    return fig, ax


def _draw_dispersion(ax, k, disp, color=None, fill=False, alpha=0.2, lw=1.0):
    x = [0, np.max(k)]
    colors = _color_list(color, disp.shape[1])
    for i, band in enumerate(disp.T):
        col = colors[i]
        if isinstance(col, Atom):
            col = col.color
        ax.plot(k, band, lw=lw, color=col)
        if fill:
            ax.fill_between(x, min(band), max(band), color=col, alpha=alpha)


def plot_dispersion(
    disp,
    labels,
    xlabel="$k$",
    ylabel="$E(k)$",
    grid="both",
    color=None,
    alpha=0.2,
    lw=1.0,
    scales=None,
    fill=False,
    ax=None,
    show=True,
):
    num_points = len(labels)
    k, ticks = _scale_xaxis(num_points, disp, scales)
    if ax is None:
        fig, ax = bandpath_subplots(ticks, labels, xlabel, ylabel, grid)
    else:
        fig = ax.get_figure()

    x = [0, np.max(k)]
    colors = _color_list(color, disp.shape[1])
    for i, band in enumerate(disp.T):
        col = colors[i]
        if isinstance(col, Atom):
            col = col.color
        ax.plot(k, band, lw=lw, color=col)
        if fill:
            ax.fill_between(x, min(band), max(band), color=col, alpha=alpha)

    fig.tight_layout()
    if show:
        plt.show()
    return ax


def disp_dos_subplots(
    ticks,
    labels,
    xlabel="$k$",
    ylabel="$E(k)$",
    doslabel="$n(E)$",
    wratio=(3, 1),
    grid="both",
):
    fig, axs = plt.subplots(1, 2, gridspec_kw={"width_ratios": wratio}, sharey="all")
    ax1, ax2 = axs
    ax1.set_xlim(0, ticks[-1])
    if xlabel:
        ax1.set_xlabel(xlabel)
    if ylabel:
        ax1.set_ylabel(ylabel)
    if doslabel:
        ax2.set_xlabel(doslabel)
    ax1.set_xticks(ticks)
    ax1.set_xticklabels(labels)
    ax2.set_xticks([0])
    if grid:
        ax1.set_axisbelow(True)
        ax1.grid(b=True, which="major", axis=grid)
        ax2.set_axisbelow(True)
        ax2.grid(b=True, which="major", axis=grid)
    return fig, axs


def plot_disp_dos(
    disp,
    dos_data,
    labels,
    xlabel="k",
    ylabel="E(k)",
    doslabel="n(E)",
    wratio=(3, 1),
    grid="both",
    color=None,
    fill=True,
    disp_alpha=0.2,
    dos_alpha=0.2,
    lw=1.0,
    scales=None,
    axs=None,
    show=True,
):
    num_points = len(labels)
    k, ticks = _scale_xaxis(num_points, disp, scales)
    if axs is None:
        fig, axs = disp_dos_subplots(
            ticks, labels, xlabel, ylabel, doslabel, wratio, grid
        )
        ax1, ax2 = axs
    else:
        ax1, ax2 = axs
        fig = ax1.get_figure()

    x = [0, np.max(k)]
    colors = _color_list(color, disp.shape[1])
    for i, band in enumerate(disp.T):
        col = colors[i]
        if isinstance(col, Atom):
            col = col.color
        ax1.plot(k, band, lw=lw, color=col)
        if fill:
            ax1.fill_between(x, min(band), max(band), color=col, alpha=disp_alpha)

    for i, band in enumerate(dos_data):
        col = colors[i]
        if isinstance(col, Atom):
            col = col.color
        bins, dos = band
        ax2.plot(dos, bins, lw=lw, color=col)
        ax2.fill_betweenx(bins, 0, dos, alpha=dos_alpha, color=col)

    ax2.set_xlim(0, ax2.get_xlim()[1])
    fig.tight_layout()
    if show:
        plt.show()
    return axs


def plot_bands(
    kgrid,
    bands,
    k_label="k",
    disp_label="E(k)",
    grid="both",
    contour_grid=False,
    bz=None,
    pi_ticks=True,
    ax=None,
    show=True,
):
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    dim = len(bands.shape) - 1
    if dim == 1:
        k = kgrid[0]
        ax.plot(k, bands.T)
        if k_label:
            ax.set_xlabel(f"${k_label}$")
        if disp_label:
            ax.set_ylabel(f"${disp_label}$")
        if grid:
            ax.grid(axis=grid)

        if pi_ticks:
            _set_piticks(ax.xaxis, num_ticks=2)
        ax.set_xlim(np.min(k), np.max(k))

        if bz is not None:
            for x in bz:
                ax.axvline(x=x, color="k")

    elif dim == 2:
        kx, ky = kgrid
        kxx, kyy = np.meshgrid(kx, ky)
        if len(bands) == 1:
            bands = bands[0]
        else:
            bands = np.sum(np.abs(bands), axis=0)

        im = ax.contourf(kxx, kyy, bands)
        ax.set_aspect("equal")
        if k_label:
            ax.set_xlabel(f"{k_label}$_x$")
            ax.set_ylabel(f"{k_label}$_y$")
        if disp_label:
            label = ""
            if disp_label:
                label = disp_label if len(bands) == 1 else f"|{disp_label}|"
            fig.colorbar(im, ax=ax, label=label)
        if grid and contour_grid:
            ax.grid(axis=grid)

        if pi_ticks:
            _set_piticks(ax.xaxis, num_ticks=2)
            _set_piticks(ax.yaxis, num_ticks=2)

        if bz is not None:
            draw_lines(ax, bz, color="k")
    else:
        raise NotImplementedError()

    fig.tight_layout()
    if show:
        plt.show()
    return ax


class DispersionPath:
    """Defines a dispersion path between high symmetry (HS) points.

    Examples
    --------
    Define a path using the add-method or preset points. To get the actual points the
    'build'-method is called:

    >>> path = DispersionPath(dim=3).add([0, 0, 0], 'Gamma').x(a=1.0).cycle()
    >>> vectors = path.build(n_sect=1000)

    Attributes
    ----------
    dim : int
    labels : list of str
    points : list of array_like
    n_sect : int
    """

    def __init__(self, dim=0):
        self.dim = dim
        self.labels = list()
        self.points = list()
        self.n_sect = 0

    @classmethod
    def chain_path(cls, a=1.0):
        return cls(dim=1).x(a).gamma().cycle()

    @classmethod
    def square_path(cls, a=1.0):
        return cls(dim=2).gamma().x(a).m(a).cycle()

    @classmethod
    def cubic_path(cls, a=1.0):
        return cls(dim=3).gamma().x(a).m(a).gamma().r(a)

    @property
    def num_points(self):
        """int: Number of HS points in the path"""
        return len(self.points)

    def add(self, point, name=""):
        """Adds a new HS point to the path

        This method returns the instance for easier path definitions.

        Parameters
        ----------
        point: array_like
            The coordinates of the HS point. If the dimension of the point is
            higher than the set dimension the point will be clipped.
        name: str, optional
            Optional name of the point. If not specified the number of the point
            is used.

        Returns
        -------
        self: DispersionPath
        """
        if not name:
            name = str(len(self.points))
        point = np.asarray(point)
        if self.dim:
            point = point[: self.dim]
        else:
            self.dim = len(point)
        self.points.append(point)
        self.labels.append(name)
        return self

    def add_points(self, points, names=None):
        """Adds multiple HS points to the path

        Parameters
        ----------
        points: array_like
            The coordinates of the HS points.
        names: list of str, optional
            Optional names of the points. If not specified the number of the point
            is used.

        Returns
        -------
        self: DispersionPath
        """
        if names is None:
            names = [""] * len(points)
        for point, name in zip(points, names):
            self.add(point, name)
        return self

    def cycle(self):
        """Adds the first point of the path.

        This method returns the instance for easier path definitions.

        Returns
        -------
        self: DispersionPath
        """
        self.points.append(self.points[0])
        self.labels.append(self.labels[0])
        return self

    def gamma(self):
        r"""DispersionPath: Adds the .math:'\Gamma=(0, 0, 0)' point to the path"""
        return self.add([0, 0, 0], r"$\Gamma$")

    def x(self, a=1.0):
        r"""DispersionPath: Adds the .math:'X=(\pi, 0, 0)' point to the path"""
        return self.add([np.pi / a, 0, 0], r"$X$")

    def m(self, a=1.0):
        r"""DispersionPath: Adds the ,math:'M=(\pi, \pi, 0)' point to the path"""
        return self.add([np.pi / a, np.pi / a, 0], r"$M$")

    def r(self, a=1.0):
        r"""DispersionPath: Adds the .math:'R=(\pi, \pi, \pi)' point to the path"""
        return self.add([np.pi / a, np.pi / a, np.pi / a], r"$R$")

    def build(self, n_sect=1000):
        """Builds the vectors defining the path between the set HS points.

        Parameters
        ----------
        n_sect: int, optional
            Number of points between each pair of HS points.

        Returns
        -------
        path: (N, D) np.ndarray
        """
        self.n_sect = n_sect
        path = np.zeros((0, self.dim))
        for p0, p1 in chain(self.points):
            path = np.append(path, np.linspace(p0, p1, num=n_sect), axis=0)
        return path

    def get_ticks(self):
        """Get the positions of the points of the last buildt path.

        Mainly used for setting ticks in plot.

        Returns
        -------
        ticks: (N) np.ndarray
        labels: (N) list
        """
        return np.arange(self.num_points) * self.n_sect, self.labels

    def edges(self):
        """Constructs the edges of the path."""
        return list(chain(self.points))

    def distances(self):
        """Computes the distances between the edges of the path."""
        dists = list()
        for p0, p1 in self.edges():
            dists.append(distance(p0, p1))
        return np.array(dists)

    def scales(self):
        """Computes the scales of the the edges of the path."""
        dists = self.distances()
        return dists / dists[0]

    def draw(self, ax, color=None, lw=1.0, **kwargs):
        lines = draw_lines(ax, self.edges(), color=color, lw=lw, **kwargs)
        return lines

    def subplots(self, xlabel="k", ylabel="E(k)", grid="both"):
        """Creates an empty matplotlib plot with configured axes for the path.

        Parameters
        ----------
        xlabel: str, optional
        ylabel: str, optional
        grid: str, optional

        Returns
        -------
        fig: plt.Figure
        ax: plt.Axis
        """
        ticks, labels = self.get_ticks()
        return bandpath_subplots(ticks, labels, xlabel, ylabel, grid)

    def plot_dispersion(self, disp, ax=None, show=True, **kwargs):
        scales = self.scales()
        return plot_dispersion(
            disp, self.labels, scales=scales, ax=ax, show=show, **kwargs
        )

    def plot_disp_dos(self, disp, dos, axs=None, show=True, **kwargs):
        scales = self.scales()
        return plot_disp_dos(
            disp, dos, self.labels, scales=scales, axs=axs, show=show, **kwargs
        )



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Spatial algorithms and data structures."""

import math
import numpy as np
import itertools
import scipy.spatial
import matplotlib.pyplot as plt
from typing import Iterable, Sequence, Union
from .utils import ArrayLike, min_dtype, chain
from .plotting import draw_points, draw_vectors, draw_lines, draw_surfaces


__all__ = [
    "distance",
    "distances",
    "interweave",
    "vindices",
    "vrange",
    "cell_size",
    "cell_volume",
    "compute_vectors",
    "KDTree",
    "VoronoiTree",
    "WignerSeitzCell",
    "rx",
    "ry",
    "rz",
    "rotate2d",
    "rotate3d",
]


def distance(r1: np.ndarray, r2: np.ndarray, decimals: int = None) -> float:
    """Calculates the euclidian distance bewteen two points.

    Parameters
    ----------
    r1 : (D, ) np.ndarray
        First input point.
    r2 : (D, ) np.ndarray
        Second input point of matching size.
    decimals: int, optional
        Decimals for rounding the output.

    Returns
    -------
    distance: float
        The distance between the input points.
    """
    dist = math.sqrt(np.sum(np.square(r1 - r2)))
    if decimals is not None:
        dist = round(dist, decimals)
    return dist


def distances(r1: ArrayLike, r2: ArrayLike, decimals: int = None) -> np.ndarray:
    """Calculates the euclidian distance between multiple points.

    Parameters
    ----------
    r1 : (N, D) array_like
        First input point.
    r2 : (N, D) array_like
        Second input point of matching size.
    decimals: int, optional
        Optional decimals to round distance to.

    Returns
    -------
    distances : (N, ) np.ndarray
        The distances for each pair of input points.
    """
    r1 = np.atleast_2d(r1)
    r2 = np.atleast_2d(r2)
    dist = np.sqrt(np.sum(np.square(r1 - r2), axis=1))
    if decimals is not None:
        dist = np.round(dist, decimals=decimals)
    return dist


def interweave(arrays: Sequence[np.ndarray]) -> np.ndarray:
    """Interweaves multiple arrays along the first axis

    Example
    -------
    >>> arr1 = np.array([[1, 1], [3, 3], [5, 5]])
    >>> arr2 = np.array([[2, 2], [4, 4], [6, 6]])
    >>> interweave([arr1, arr2])
    array([[1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6]])

    Parameters
    ----------
    arrays: (M) Sequence of (N, ...) array_like
        The input arrays to interwave. The shape of all arrays must match.

    Returns
    -------
    interweaved: (M*N, ....) np.ndarray
    """
    shape = list(arrays[0].shape)
    shape[0] = sum(x.shape[0] for x in arrays)
    result = np.empty(shape, dtype=arrays[0].dtype)
    n = len(arrays)
    for i, arr in enumerate(arrays):
        result[i::n] = arr
    return result


def vindices(
    limits: Iterable[Sequence[int]],
    sort_axis: int = 0,
    dtype: Union[int, str, np.dtype] = None,
) -> np.ndarray:
    """Return an array representing the indices of a d-dimensional grid.

    Parameters
    ----------
    limits: (D, 2) array_like
        The limits of the indices for each axis.
    sort_axis: int, optional
        Optional axis that is used to sort indices.
    dtype: int or str or np.dtype, optional
        Optional data-type for storing the lattice indices. By default the given limits
        are checked to determine the smallest possible data-type.

    Returns
    -------
    vectors: (N, D) np.ndarray
    """
    if dtype is None:
        dtype = min_dtype(limits, signed=True)
    limits = np.asarray(limits)
    dim = limits.shape[0]

    # Create meshgrid reshape grid to array of indices

    # version 1:
    # axis = np.meshgrid(*(np.arange(*lim, dtype=dtype) for lim in limits))
    # nvecs = np.asarray([np.asarray(a).flatten("F") for a in axis]).T

    # version 2:
    # slices = [slice(lim[0], lim[1], 1) for lim in limits]
    # nvecs = np.mgrid[slices].astype(dtype).reshape(dim, -1).T

    # version 3:
    size = limits[:, 1] - limits[:, 0]
    nvecs = np.indices(size, dtype=dtype).reshape(dim, -1).T + limits[:, 0]

    # Optionally sort indices along given axis
    if sort_axis is not None:
        nvecs = nvecs[np.lexsort(nvecs.T[[sort_axis]])]

    return nvecs


# noinspection PyIncorrectDocstring
def vrange(
    start=None,
    *args,
    dtype: Union[int, str, np.dtype] = None,
    sort_axis: int = 0,
    **kwargs,
) -> np.ndarray:
    """Return evenly spaced vectors within a given interval.

    Parameters
    ----------
    start: array_like, optional
        The starting value of the interval. The interval includes this value.
        The default start value is 0.
    stop: array_like
        The end value of the interval.
    step: array_like, optional
        Spacing between values. If `start` and `stop` are sequences and the `step`
        is a scalar the given step size is used for all dimensions of the vectors.
        The default step size is 1.
    sort_axis: int, optional
        Optional axis that is used to sort indices.
    dtype: dtype, optional
        The type of the output array.  If `dtype` is not given, infer the data
        type from the other input arguments.

    Returns
    -------
    vectors: (N, D) np.ndarray
    """
    # parse arguments
    if len(args) == 0:
        stop = start
        start = np.zeros_like(stop)
        step = kwargs.get("step", 1.0)
    elif len(args) == 1:
        stop = args[0]
        step = kwargs.get("step", 1.0)
    else:
        stop, step = args

    start = np.atleast_1d(start)
    stop = np.atleast_1d(stop)
    if step is None:
        step = np.ones_like(start)
    elif not hasattr(step, "__len__"):
        step = np.ones_like(start) * step

    # Create grid and reshape to array of vectors
    slices = [slice(i, f, s) for i, f, s in zip(start, stop, step)]
    array = np.mgrid[slices].reshape(len(slices), -1).T
    # Optionally sort array along given axis
    if sort_axis is not None:
        array = array[np.lexsort(array.T[[sort_axis]])]

    return array if dtype is None else array.astype(dtype)


def cell_size(vectors: ArrayLike) -> np.ndarray:
    """Computes the shape of the box spawned by the given vectors.

    Parameters
    ----------
    vectors: array_like
        The basis vectors defining the cell.

    Returns
    -------
    size: np.ndarray
    """
    max_values = np.max(vectors, axis=0)
    min_values = np.min(vectors, axis=0)
    min_values[min_values > 0] = 0
    return max_values - min_values


def cell_volume(vectors: ArrayLike) -> float:
    r"""Computes the volume of the unit cell defined by the primitive vectors.

    The volume of the unit-cell in two and three dimensions is defined by

    .. math::
        V_{2d} = \abs{a_1 \cross a_2}, \quad V_{3d} = a_1 \cdot \abs{a_2 \cross a_3}

    For higher dimensions the volume is computed using the determinant:

    .. math::
        V_{d} = \sqrt{\det{A A^T}}

    where :math:`A` is the array of vectors.

    Parameters
    ----------
    vectors: array_like
        The basis vectors defining the cell.

    Returns
    -------
    vol: float
        The volume of the unit cell.
    """
    dim = len(vectors)
    if dim == 1:
        v = float(vectors)
    elif dim == 2:
        v = np.cross(vectors[0], vectors[1])
    elif dim == 3:
        cross = np.cross(vectors[1], vectors[2])
        v = np.dot(vectors[0], cross)
    else:
        v = np.sqrt(np.linalg.det(np.dot(vectors.T, vectors)))
    return abs(v)


def compute_vectors(
    a: float,
    b: float = None,
    c: float = None,
    alpha: float = None,
    beta: float = None,
    gamma: float = None,
    decimals: int = 0,
) -> np.ndarray:
    """Computes lattice vectors by the lengths and angles."""
    if b is None and c is None:
        vectors = [a]
    elif c is None:
        alpha = np.deg2rad(alpha)
        ax = a
        bx = b * np.cos(alpha)
        by = b * np.sin(alpha)
        vectors = np.array([[ax, 0], [bx, by]])
    else:
        alpha = np.deg2rad(alpha)
        beta = np.deg2rad(beta)
        gamma = np.deg2rad(gamma)
        ax = a
        bx = b * np.cos(gamma)
        by = b * np.sin(gamma)
        cx = c * np.cos(beta)
        cy = (abs(c) * abs(b) * np.cos(alpha) - bx * cx) / by
        cz = np.sqrt(c**2 - cx**2 - cy**2)
        vectors = np.array([[ax, 0, 0], [bx, by, 0], [cx, cy, cz]])
    if decimals:
        vectors = np.round(vectors, decimals=decimals)
    return vectors


# noinspection PyUnresolvedReferences
class KDTree(scipy.spatial.cKDTree):
    """Simple wrapper of scipy's cKTree with global query settings."""

    def __init__(self, points, k=1, max_dist=np.inf, eps=0.0, p=2, boxsize=None):
        super().__init__(points, boxsize=boxsize)
        self.max_dist = max_dist
        self.k = k
        self.p = p
        self.eps = eps

    def __repr__(self):
        return f"{self.__class__.__name__}(k: {self.k}, p: {self.p}, eps: {self.eps})"

    def query_ball_point(self, x, r, *_):
        return super().query_ball_point(x, r, self.p, self.eps)

    def query_ball_tree(self, other: Union["KDTree", scipy.spatial.KDTree], r, *_):
        return super().query_ball_tree(other, r, self.p, self.eps)

    def query_pairs(self, r, *_):
        return super().query_pairs(r, self.p, self.eps)

    def query(
        self, x=None, num_jobs=1, decimals=None, include_zero=False, compact=True, *_
    ):
        x = self.data if x is None else x
        dists, neighbors = super().query(
            x, self.k, self.eps, self.p, self.max_dist, num_jobs
        )

        # Remove zero-distance neighbors and convert dtype
        if not include_zero and np.all(dists[:, 0] == 0):
            dists = dists[:, 1:]
            neighbors = neighbors[:, 1:]
        neighbors = neighbors.astype(min_dtype(self.n, signed=False))

        # Remove neighbors with distance larger than max_dist
        if self.max_dist < np.inf:
            invalid = dists > self.max_dist
            neighbors[invalid] = self.n
            dists[invalid] = np.inf

        # Remove all invalid columns
        if compact:
            mask = np.any(dists != np.inf, axis=0)
            neighbors = neighbors[:, mask]
            dists = dists[:, mask]

        # Round distances
        if decimals is not None:
            dists = np.round(dists, decimals=decimals)

        return neighbors, dists


class VoronoiTree:
    def __init__(self, points):
        points = np.asarray(points)
        dim = points.shape[1]
        edges = list()
        if dim == 1:
            vertices = points / 2
            idx = np.where((vertices == np.zeros(vertices.shape[1])).all(axis=1))[0]
            vertices = np.delete(vertices, idx)
            vertices = np.atleast_2d(vertices).T
        else:
            vor = scipy.spatial.Voronoi(points)
            # Save only finite vertices
            vertices = vor.vertices  # noqa
            for pointidx, simplex in zip(vor.ridge_points, vor.ridge_vertices):  # noqa
                simplex = np.asarray(simplex)
                if np.all(simplex >= 0):
                    edges.append(simplex)

        self.dim = dim
        self.points = points
        self.edges = edges
        self.vertices = vertices
        self.tree = scipy.spatial.KDTree(points)
        self.origin = self.query(np.zeros(dim))

    def query(self, x, k=1, eps=0):
        return self.tree.query(x, k, eps)  # noqa

    def draw(
        self,
        ax=None,
        color="C0",
        size=3,
        lw=1,
        alpha=0.15,
        point_color="k",
        point_size=3,
        draw_data=True,
        points=True,
        draw=True,
        fill=True,
    ):  # pragma: no cover

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d" if self.dim == 3 else None)

        if draw_data:
            draw_points(ax, self.points, size=point_size, color=point_color)
            if self.dim > 1:
                draw_vectors(ax, self.points, lw=0.5, color=point_color)

        if points:
            draw_points(ax, self.vertices, size=size, color=color)

        if self.dim == 2 and draw:
            segments = np.array([self.vertices[i] for i in self.edges])
            draw_lines(ax, segments, color=color, lw=lw)
        elif self.dim == 3:
            if draw:
                segments = np.array(
                    [self.vertices[np.append(i, i[0])] for i in self.edges]
                )
                draw_lines(ax, segments, color=color, lw=lw)
            if fill:
                surfaces = np.array([self.vertices[i] for i in self.edges])
                draw_surfaces(ax, surfaces, color=color, alpha=alpha)

        if self.dim == 3:
            ax.set_aspect("equal")
        else:
            ax.set_aspect("equal", "box")

        return ax

    def __repr__(self):
        return f"{self.__class__.__name__}(vertices: {len(self.vertices)})"

    def __str__(self):
        return f"vertices:\n{self.vertices}\n" f"egdes:\n{self.edges}"


class WignerSeitzCell(VoronoiTree):
    def __init__(self, points):
        super().__init__(points)
        self._root = self.query(np.zeros(self.dim))[1]

    @property
    def limits(self):
        return np.array(
            [np.min(self.vertices, axis=0), np.max(self.vertices, axis=0)]
        ).T

    @property
    def size(self):
        return self.limits[1] - self.limits[0]

    def check(self, points):
        cells = np.asarray(self.query(points)[1])
        return cells == self._root

    def arange(self, steps, offset=0.0):
        limits = self.limits * (1 + offset)
        steps = [steps] * self.dim if not hasattr(steps, "__len__") else steps
        return [np.arange(*lims, step=step) for lims, step in zip(limits, steps)]

    def linspace(self, nums, offset=0.0, endpoint=False):
        limits = self.limits * (1 + offset)
        nums = [nums] * self.dim if not hasattr(nums, "__len__") else nums
        values = list()
        for lims, num in zip(limits, nums):
            values.append(np.linspace(*lims, num=num, endpoint=endpoint))
        return values

    def meshgrid(self, nums=None, steps=None, offset=0.0, check=True, endpoint=False):
        if nums is not None:
            grid = np.array(np.meshgrid(*self.linspace(nums, offset, endpoint)))
        elif steps is not None:
            grid = np.array(np.meshgrid(*self.arange(steps, offset)))
        else:
            raise ValueError(
                "Either the number of points or the step size must be specified"
            )

        if check:
            lengths = grid.shape[1:]
            dims = range(len(lengths))
            for item in itertools.product(*[range(n) for n in lengths]):
                point = np.array([grid[d][item] for d in dims])
                if not self.check(point):
                    for d in dims:
                        grid[d][item] = np.nan
        return grid

    def symmetry_points(self):
        origin = np.zeros(self.dim)
        corners = self.vertices.copy()
        face_centers = None
        if self.dim == 1:
            return origin, corners, None, None
        elif self.dim == 2:
            edge_centers = np.zeros((len(self.edges), 2))
            for i, simplex in enumerate(self.edges):
                p1, p2 = self.vertices[simplex]
                edge_centers[i] = p1 + (p2 - p1) / 2
        elif self.dim == 3:
            edge_centers = list()
            face_centers = list()
            for i, simplex in enumerate(self.edges):
                edges = self.vertices[simplex]
                # compute face centers
                face_centers.append(np.mean(edges, axis=0))
                # compute edge centers
                for p1, p2 in chain(edges, cycle=True):
                    edge_centers.append(p1 + (p2 - p1) / 2)
            edge_centers = np.asarray(edge_centers)
            face_centers = np.asarray(face_centers)
        else:
            raise NotImplementedError()
        return origin, corners, edge_centers, face_centers


def rx(theta: float) -> np.ndarray:
    """X-Rotation matrix."""
    sin, cos = np.sin(theta), np.cos(theta)
    return np.array([[1, 0, 0], [0, cos, -sin], [0, sin, cos]])


def ry(theta: float) -> np.ndarray:
    """Y-Rotation matrix."""
    sin, cos = np.sin(theta), np.cos(theta)
    return np.array([[cos, 0, sin], [0, 1, 0], [-sin, 0, +cos]])


def rz(theta: float) -> np.ndarray:
    """Z-Rotation matrix."""
    sin, cos = np.sin(theta), np.cos(theta)
    return np.array([[cos, -sin, 0], [sin, cos, 0], [0, 0, 1]])


def rot(
    thetax: float = 0.0, thetay: float = 0.0, thetaz: float = 0.0
):  # pragma: no cover
    """General rotation matrix"""
    r = np.eye(3)
    if thetaz:
        r = np.dot(r, rz(thetaz))
    if thetay:
        r = np.dot(r, ry(thetay))
    if thetax:
        r = np.dot(r, rz(thetax))
    return r


def rotate2d(a, theta, degree=True):  # pragma: no cover
    """Applies the z-rotation matrix to a 2D point"""
    if degree:
        theta = np.deg2rad(theta)
    return np.dot(a, rz(theta)[:2, :2])


def rotate3d(a, thetax=0.0, thetay=0.0, thetaz=0.0, degree=True):  # pragma: no cover
    """Applies the general rotation matrix to a 3D point"""
    if degree:
        thetax = np.deg2rad(thetax)
        thetay = np.deg2rad(thetay)
        thetaz = np.deg2rad(thetaz)
    return np.dot(a, rot(thetax, thetay, thetaz))



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

import numpy as np
from pytest import mark
from numpy.testing import assert_array_equal
from lattpy import utils


@mark.parametrize(
    "a, signed, result",
    [
        (+127, False, np.uint8),
        (-127, False, np.int8),
        (+127, True, np.int8),
        (+128, False, np.uint8),
        (-128, False, np.int8),
        (+128, True, np.int16),
        ([-128, 127], False, np.int8),
        ([-128, 128], False, np.int16),
        ([-129, 127], False, np.int16),
        ([+30, 127], False, np.uint8),
    ],
)
def test_min_dtype(a, signed, result):
    assert utils.min_dtype(a, signed) == result


@mark.parametrize(
    "items, cycle, result",
    [
        ([0, 1, 2], False, [[0, 1], [1, 2]]),
        ([0, 1, 2], True, [[0, 1], [1, 2], [2, 0]]),
        (["0", "1", "2"], False, [["0", "1"], ["1", "2"]]),
        (["0", "1", "2"], True, [["0", "1"], ["1", "2"], ["2", "0"]]),
    ],
)
def test_chain(items, cycle, result):
    assert_array_equal(utils.chain(items, cycle), result)
